(*
   This file is part of the Anoq SML Basis project version 0.8.2.
   This  project implements Anoq's proposals for some extensions to the
   Standard ML Basis Library, which will make the Basis Library interfaces
   more consistent.

   Copyright (C) 2003-2009 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation;
   version 2.1 of the License.

   As a special exception, when making a "work that uses the library",
   you may compile this library directly into that work, without making
   it possible to relink that work with a modified version of this library.
   This is a relaxation of some of the requirements in section 6 of the
   GNU Lesser General Public License version 2.1.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License version 2.1 for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information:
   http://www.HardcoreProcessing.com/contact

   You should have received a copy of the GNU Lesser General Public
   License version 2.1 (WITHOUT the option of using a later version)
   along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*)

structure List = (* Define a new and better List module :) *)
  struct
    open List (* The current implementation of List *)

    type index = int
    type dimensions = int
    type 'a t = 'a list

    (* The functions nth and length could be deprecated *)
    val sub = List.nth
    val dimensions = List.length
    val count = dimensions
    val append = op@

    (* The functions take and drop could be deprecated *)
    (* Substring.splitAt raises Subscript
       Substring.triml/r do not.
       List.take/drop do.
       Inconsistent? I think that all these functions should raise Subscript. *)
    fun triml dim l =
          List.drop (l, dim)

    fun trimr dim l =
          List.take (l, List.length l - dim)

    fun keepl dim l =
          List.take (l, dim)

    fun keepr dim l =
          List.drop (l, List.length l - dim)

    fun empty () = nil
    fun create (dim, e) = List.tabulate (dim, fn _ => e)
    fun toList l = l
    fun fromList l = l

    fun tabulatel (dim, f, acc) =
          let
            fun tabulatelF i acc result =
                  if i < dim then
                    let
                      val (e, a) = f (i, acc)
                    in
                      tabulatelF (i + 1) a (e::result)
                    end
                  else
                    (result, acc)
            
            val (l, a) = tabulatelF 0 acc nil
          in
            (rev l, a)
          end

    fun tabulater (dim, f, acc) =
          let
            fun tabulaterF i acc result =
                  if i >= 0 then
                    let
                      val (e, a) = f (i, acc)
                    in
                      tabulaterF (i - 1) a (e::result)
                    end
                  else
                    (result, acc)
          in
            tabulaterF (dim - 1) acc nil
          end
    
    fun appi f l =
          let
            fun appiF cnt (e::es) =
                  (f (cnt, e) : unit;
                   appiF (cnt + 1) es)
              | appiF _ nil =
                  ()
          in
            appiF 0 l
          end

    fun mapi f a =
          let
            fun mapiF i (x::xs) =
                  (f (i, x)) :: (mapiF (i + 1) xs)
              | mapiF _ nil =
                  nil
          in
            mapiF 0 a
          end

    fun existsi p l =
          let
            fun existsiF i (x::xs) =
                  if p (i, x) then
                    true
                  else
                    existsiF (i + 1) xs
              | existsiF _ nil =
                  false
          in
            existsiF 0 l
          end

    fun findi p l =
          let
            fun findiF i (x::xs) =
                  if p (i, x) then
                    SOME (i, x)
                  else
                    findiF (i + 1) xs
              | findiF _ nil =
                  NONE
          in
            findiF 0 l
          end

    fun foldli f init l =
          #2 (foldl (fn (e, (i, acc)) => (i + 1, f (i, e, acc))) (0, init) l)

    fun foldri f init l =
          let
            fun foldriF i (x::xs) =
                  f (i, x, foldriF (i + 1) xs)
              | foldriF _ nil =
                  init
          in
            foldriF 0 l
          end

    fun mapli f init l =
          let
            fun mapliF i acc (x::xs) =
                  let
                    val (e, acc') = f(i, x, acc)
                    val (l, accResult) = mapliF (i + 1) acc' xs
                  in
                    (e::l, accResult)
                  end
              | mapliF _ acc nil =
                  (nil, acc)
          in
            mapliF 0 init l
          end

    fun mapri f init l =
          let
            fun mapriF i (x::xs) =
                  let
                    val (l, acc) = mapriF (i + 1) xs
                    val (e, acc') = f(i, x, acc)
                  in
                    (e::l, acc')
                  end
              | mapriF _ nil =
                  (nil, init)
          in
            mapriF 0 l
          end

    fun mapl f init l =
        mapli (fn (i, e, acc) => f(e, acc)) init l

    fun mapr f init l =
        mapri (fn (i, e, acc) => f(e, acc)) init l

    fun foldsepl sep f init (x::xs) =
          foldl (fn (e, acc) => f (e, sep acc)) (f (x, init)) xs
      | foldsepl _ _ init nil =
          init

    fun foldsepr sep f init l =
          foldsepl sep f init (rev l)

    fun foldsepli sep f init l =
          let
            fun foldF (i, e, acc) =
                  if i > 0 then
                    f (i, e, sep acc)
                  else
                    f (i, e, acc)
          in
            foldli foldF init l
          end

    fun foldsepri sep f init l =
          let
            fun foldF (i, e, acc) =
                  if i > 0 then
                    sep (f (i, e, acc))
                  else
                    f (i, e, acc)
          in
            foldri foldF init l
          end

    (* FIXME: Not tested *)
    fun slice (l, start, SOME len) =
        List.take(List.drop(l, start), len)
      | slice (l, start, NONE) =
        List.drop(l, start)
    
    fun lookup (x::xs) index =
          if index > 0 then
            (* Notice: Optimized with the most common case first *)
            lookup xs (index - 1)
          else if index = 0 then
            SOME x
          else
            NONE
      | lookup nil _ =
          NONE

    fun insert index v (e::es) =
          if index > 0 then
            (* Notice: Optimized with the most common case first *)
            e :: (insert (index - 1) v es)
          else if index = 0 then
            v::e::es
          else
            raise Subscript
      | insert index v nil =
          if index = 0 then
            [v]
          else
            raise Subscript
          
    fun delete index (e::es) =
          if index > 0 then
            (* Notice: Optimized with the most common case first *)
            e :: (delete (index - 1) es)
          else if index = 0 then
            es
          else
            raise Subscript
      | delete index nil =
          raise Subscript
      
    fun substitute index f (e::es) =
          if index > 0 then
            (* Notice: Optimized with the most common case first *)
            e :: (substitute (index - 1) f es)
          else if index = 0 then
            (f e)::es
          else
            raise Subscript
      | substitute index f nil =
          raise Subscript

    fun filteri p l =
          let
            fun filteriF i (x::xs) =
                  if p (i, x) then
                    x :: (filteriF (i + 1) xs)
                  else
                    filteriF (i + 1) xs
              | filteriF _ nil =
                  nil
          in
            filteriF 0 l
          end

    fun splitAt (x::xs, i) =
          if i > 0 then
            let
              val (l, r) = splitAt (xs, i - 1)
            in
              (x::l, r)
            end
          else if i = 0 then
            (nil, x::xs)
          else
            raise Subscript
      | splitAt (nil, i) =
          if i = 0 then
            (nil, nil)
          else
            raise Subscript
      
    fun splitl p (x::xs) =
          if p x then
            let
              val (l, r) = splitl p xs
            in
              (x::l, r)
            end
          else
            (nil, x::xs)
      | splitl p nil =
          (nil, nil)

    fun splitli p l =
          let
            fun splitliF i (x::xs) =
                  if p (i, x) then
                    let
                      val (l, r) = splitliF (i + 1) xs
                    in
                      (x::l, r)
                    end
                  else
                    (nil, x::xs)
              | splitliF _ nil =
                  (nil, nil)
          in
            splitliF 0 l
          end

    fun splitri p l =
          let
            fun splitriF i (x::xs) =
                  (case splitriF (i + 1) xs of
                     (NONE, r) =>
                       if p (i, x) then
                         (NONE, x::r)
                       else
                         (SOME [x], r)
                   | (SOME l, r) =>
                       (SOME (x::l), r)
                  )
              | splitriF _ nil =
                  (NONE, nil)
          in
            case splitriF 0 l of
              (NONE, r) =>
                (nil, r)
            | (SOME l, r) =>
                (l, r)
          end

    fun splitr p l =
          splitri (fn (_, x) => p x) l

    (* The take and drop functions can always be implemented by split. *)
    fun takel p l =
          #1 (splitl p l)

    fun takeli p l =
          #1 (splitli p l)

    fun taker p l =
          #2 (splitr p l)

    fun takeri p l =
          #2 (splitri p l)
    
    fun dropl p l =
          #2 (splitl p l)

    fun dropli p l =
          #2 (splitli p l)

    fun dropr p l =
          #1 (splitr p l)

    fun dropri p l =
          #1 (splitri p l)
  end
