(*
   This file is part of the Anoq SML Basis project version 0.8.2.
   This  project implements Anoq's proposals for some extensions to the
   Standard ML Basis Library, which will make the Basis Library interfaces
   more consistent.

   Copyright (C) 2003-2009 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation;
   version 2.1 of the License.

   As a special exception, when making a "work that uses the library",
   you may compile this library directly into that work, without making
   it possible to relink that work with a modified version of this library.
   This is a relaxation of some of the requirements in section 6 of the
   GNU Lesser General Public License version 2.1.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License version 2.1 for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information:
   http://www.HardcoreProcessing.com/contact

   You should have received a copy of the GNU Lesser General Public
   License version 2.1 (WITHOUT the option of using a later version)
   along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*)

(* This is supposed to match ALL monomorphic collection modules in the
   Standard ML Basis library.
   I.e. Word8Vector, String, IntVector etc.
   It should also match most other monomorphic collections. *)
signature MONO_COLLECTION_NON_REDUCIBLE =
  sig
    type index
    type dimensions
    type t
    type elem
    
    (* Construction operations *)
    val empty      : unit -> t
    val create     : dimensions * elem -> t
    val tabulate   : dimensions * (index -> elem) -> t
    
    (* tabulatel/r are new and badly missing in SML Basis. Could be curried? *)
    val tabulatel  : dimensions * (index * 'b -> elem * 'b) * 'b -> t * 'b
    val tabulater  : dimensions * (index * 'b -> elem * 'b) * 'b -> t * 'b
    
    (* Read operations *)
    val sub        : t * index -> elem
    (* Dynamic array and dictionary are really good examples of
       difference between dimensions and count.
       count returns exact number of elements. dimensions returns what
       makes sense for data structure dimensions, e.g. the list of all
       used keys in a dictionary or the allocated array size of a dynamic
       array. The type dimensions must make sense for tabulate etc. *)
    val count      : t -> int
    val dimensions : t -> dimensions

    (* Read-only traversal operations *)
    val foldl     : (elem * 'b -> 'b) -> 'b -> t -> 'b
    val foldli    : (index * elem * 'b -> 'b) -> 'b -> t -> 'b
    val foldr     : (elem * 'b -> 'b) -> 'b -> t -> 'b
    val foldri    : (index * elem * 'b -> 'b) -> 'b -> t -> 'b

    val foldsepl  : ('b -> 'b) -> (elem * 'b -> 'b) -> 'b -> t -> 'b
    val foldsepli : ('b -> 'b) -> (index * elem * 'b -> 'b) -> 'b -> t -> 'b
    val foldsepr  : ('b -> 'b) -> (elem * 'b -> 'b) -> 'b -> t -> 'b
    val foldsepri : ('b -> 'b) -> (index * elem * 'b -> 'b) -> 'b -> t -> 'b
    
    val exists    : (elem -> bool) -> t -> bool
    val existsi   : (index * elem -> bool) -> t -> bool
    val find      : (elem -> bool) -> t -> elem option
    val findi     : (index * elem -> bool) -> t -> (index * elem) option
    (* FIXME: val all       : (elem -> bool) -> t -> bool *)
    (* FIXME: val collate   : (elem * elem -> order) -> t * t -> order *)
    val app       : (elem -> unit) -> t -> unit
    val appi      : (index * elem -> unit) -> t -> unit
    
    (* Functional operations *)
    val rev       : t -> t
    val append    : t * t -> t
    val concat    : t list -> t
    (* FIXME: val fupdate   : t * index * elem -> t *)
    val map       : (elem -> elem) -> t -> t
    val mapi      : ((index * elem) -> elem) -> t -> t

    (* map{l|r}[i] are new and badly missing in Standard ML Basis. *)
    val mapli     : ((index * elem * 'b) -> (elem * 'b)) -> 'b -> t -> (t * 'b)
    val mapri     : ((index * elem * 'b) -> (elem * 'b)) -> 'b -> t -> (t * 'b)
    val mapl      : ((elem * 'b) -> (elem * 'b)) -> 'b -> t -> (t * 'b)
    val mapr      : ((elem * 'b) -> (elem * 'b)) -> 'b -> t -> (t * 'b)

    (* Only the String module already has the extract function. *)
    val slice     : t * index * dimensions option -> t

    (* Conversion to and from lists *)
    val toList    : t -> elem list
    val fromList  : elem list -> t

    (* Dictionary functions, which do not alter collection dimensions. *)
    val lookup    : t -> index -> elem option
    val substitute : index -> (elem -> elem) -> t -> t
    (* The name update is used by some of the now obsolete Basis signatures,
       and we already use the names fupdate and iupdate,
       so we call it substitute. *)

    (* Dictionary function. insert may enlarge dimensions for some
       collections. This may not make sense for all kinds of collections. *)
    val insert    : index -> elem -> t -> t
  end
