(*
   This file is part of the Anoq SML Basis project version 0.8.2.
   This  project implements Anoq's proposals for some extensions to the
   Standard ML Basis Library, which will make the Basis Library interfaces
   more consistent.

   Copyright (C) 2003-2009 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation;
   version 2.1 of the License.

   As a special exception, when making a "work that uses the library",
   you may compile this library directly into that work, without making
   it possible to relink that work with a modified version of this library.
   This is a relaxation of some of the requirements in section 6 of the
   GNU Lesser General Public License version 2.1.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License version 2.1 for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information:
   http://www.HardcoreProcessing.com/contact

   You should have received a copy of the GNU Lesser General Public
   License version 2.1 (WITHOUT the option of using a later version)
   along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*)

structure String = (* Define a new and better String/CharVector module *)
  struct
    open String (* The current implementation of String (for compatibility) *)

    type index = int
    type dimensions = int
    type t = string
    type elem = char
    type vector = t (* CharVector compatibility *)

    (* Notice: size, explode and implode could all be deprecated *)
    val dimensions = size
    val count      = dimensions
    val toList     = explode
    val fromList   = implode (* Also the same as CharVector.fromList *)
    val append     = String.^
    val slice      = extract

    (* We use the native CharVector implementations, whereever possible *)
    val tabulate = CharVector.tabulate
    val foldl    = CharVector.foldl
    val foldr    = CharVector.foldr
    val map      = CharVector.map
    val app      = CharVector.app
    val update   = AnoqCharVector2002.update (* FIXME: Should only be in CharVector for backwards compatibility *)
    val all      = AnoqCharVector2002.all
    val foldli   = AnoqCharVector2002.foldli
    val foldri   = AnoqCharVector2002.foldri
    val mapi     = AnoqCharVector2002.mapi
    val appi     = AnoqCharVector2002.appi

    (* Backwards compatibility with CharVector *)
    val maxLen = CharVector.maxLen (* Probably the same as String.maxSize *)
    val length = CharVector.length (* Use dimensions or count *)

    (* We use the native CharVector implementations *)
    (*
    fun foldli f acc str =
          let
            val d = dimensions str
            
            fun foldliIter i acc =
                  if Int.< (i, d) then
                    foldliIter (i + 1) (f (i, sub (str, i), acc))
                  else
                    acc
          in
            foldliIter 0 acc
          end

    fun foldl f acc str =
          foldli (fn (_, e, acc) => f (e, acc)) acc str

    fun foldri f acc str =
          let
            val d = dimensions str
            
            fun foldriIter i acc =
                  if Int.>= (i, 0) then
                    foldriIter (i - 1) (f (i, sub (str, i), acc))
                  else
                    acc
          in
            foldriIter (d - 1) acc
          end

    fun foldr f acc str =
          foldri (fn (_, e, acc) => f (e, acc)) acc str
    *)

    fun mapli f init str =
          let
            fun foldF (i, e, (resultarr, acc)) =
                 let
                   val (e', acc') = f (i, e, acc)
                 in
                   (e'::resultarr, acc')
                 end

            val (l, acc') = foldli foldF (nil, init) str
          in
            (fromList (List.rev l), acc')
          end

    fun mapri f init str =
          let
            fun foldF (i, e, (resultarr, acc)) =
                 let
                   val (e', acc') = f (i, e, acc)
                 in
                   (e'::resultarr, acc')
                 end

            val (l, acc') = foldri foldF (nil, init) str
          in
            (fromList l, acc')
          end

    fun mapl f init str =
          mapli (fn (_, e, acc) => f (e, acc)) init str

    fun mapr f init str =
          mapri (fn (_, e, acc) => f (e, acc)) init str


    fun foldsepli sep f init l =
          let
            fun foldF (i, e, acc) =
                  if Int.>(i, 0) then
                    f (i, e, sep acc)
                  else
                    f (i, e, acc)
          in
            foldli foldF init l
          end

    fun foldsepri sep f init l =
          let
            fun foldF (i, e, acc) =
                  if Int.>(i, 0) then
                    sep (f (i, e, acc))
                  else
                    f (i, e, acc)
          in
            foldri foldF init l
          end

    fun foldsepl sep f acc str =
          foldsepli sep (fn (_, e, acc) => f (e, acc)) acc str

    fun foldsepr sep f acc str =
          foldsepri sep (fn (_, e, acc) => f (e, acc)) acc str

    (* We use the native CharVector implementations *)
    (*
    fun mapi f str =
          fromList
            (List.rev
               (foldli (fn (i, e, acc) => (f (i, e))::acc) nil str)
            )

    fun appi f str =
          foldli (fn (i, e, ()) => f (i, e)) () str

    fun app f str =
          foldl (fn (e, ()) => f e) () str

    fun tabulate args =
          fromList
            (List.tabulate args)
    *)
    
    (* The operations from here are implemented from the basic operations. *)
    fun empty () =
          tabulate (0, fn _ => raise Subscript)
          
    fun create (dim, e) =
          tabulate (dim, fn _ => e)

    fun tabulatel (dim, f, init) =
          let
            val (l, result) = List.tabulatel (dim, f, init)
          in
            (fromList l, result)
          end

    fun tabulater (dim, f, init) =
          let
            val (l, result) = List.tabulater (dim, f, init)
          in
            (fromList l, result)
          end

    fun rev vec =
          fromList (foldl op:: nil vec)

    local
      fun findF f (e, NONE) =
            if f e then
              SOME e
            else
              NONE
        | findF f (_, acc) =
            (* acc is: SOME e *)
            acc

      fun findiF f (i, e, NONE) =
            if f (i, e) then
              SOME (i, e)
            else
              NONE
        | findiF f (_, _, acc) =
            (* acc is: SOME e *)
            acc
    in
      fun find f vec =
            foldl (findF f) NONE vec

      fun findi f vec =
            foldli (findiF f) NONE vec
    end

    fun exists f vec =
          foldl (fn (e, acc) => acc orelse (f e)) false vec

    fun existsi f vec =
          foldli (fn (i, e, acc) => acc orelse (f (i, e))) false vec

    fun lookup vec index =
          (* It really ought to be sub which is implemented in terms
             of lookup, but we already have sub, so I haven't bothered. *)
          if Int.>= (index, 0) andalso Int.< (index, dimensions vec) then
            SOME (sub (vec, index))
          else
            NONE

    fun insert index v vec =
          tabulate
            (dimensions vec + 1,
             fn i => if Int.< (i, index) then
                       sub (vec, i)
                     else if i = index then
                       v
                     else
                       sub (vec, i - 1)
            )

    fun delete index vec =
          if Int.>= (index, 0) then
            tabulate
               (dimensions vec - 1,
                fn i => if Int.< (i, index) then
                          sub (vec, i)
                        else
                          sub (vec, i + 1)
               )
          else
             (* If index < 0 then the tabulated function is not called,
                so we have to explicitly make this range check *)
             raise Subscript

    fun substitute index f vec =
          mapi (fn (i, v) => if i = index then f v else v) vec

    fun filteri p l =
          fromList
            (List.rev
              (foldli
                 (fn (i, e, acc) => if p (i, e) then e :: acc else acc)
                 nil l
              )
            )

    fun filter p l =
          fromList
            (List.rev
              (foldl
                 (fn (e, acc) => if p e then e :: acc else acc)
                 nil l
              )
            )

    fun triml dim vec =
          let
            val d = dimensions vec
          in
            if Int.>= (dim, 0) andalso Int.<= (dim, d) then
              tabulate (dimensions vec - dim, fn i => sub (vec, dim + i))
            else
              raise Subscript
          end

    fun trimr dim vec =
          let
            val d = dimensions vec
          in
            if Int.>= (dim, 0) andalso Int.<= (dim, d) then
              tabulate (dimensions vec - dim, fn i => sub (vec, i))
            else
              raise Subscript
          end

    fun keepl dim vec =
          let
            val d = dimensions vec
          in
            if Int.>= (dim, 0) andalso Int.<= (dim, d) then
              tabulate (dim, fn i => sub (vec, i))
            else
              raise Subscript
          end

    fun keepr dim vec =
          let
            val d = dimensions vec
            val offset = d - dim
          in
            if Int.>= (dim, 0) andalso Int.<= (dim, d) then
              tabulate (dim, fn i => sub (vec, offset + i))
            else
              raise Subscript
          end

    fun splitAt (vec, i) =
          (keepl i vec,
           triml i vec)

    fun splitl p l =
          let
            fun splitlF (e, (left, right, continue)) =
                  if continue andalso p e then
                    (e::left, right, true)
                  else
                    (left, e::right, false)

            val (left, right, _) =
                  foldl splitlF (nil, nil, true) l
          in
            (fromList (List.rev left),
             fromList (List.rev right))
          end

    fun splitli p l =
          let
            fun splitliF (i, e, (left, right, continue)) =
                  if continue andalso p (i, e) then
                    (e::left, right, true)
                  else
                    (left, e::right, false)

            val (left, right, _) =
                  foldli splitliF (nil, nil, true) l
          in
            (fromList (List.rev left),
             fromList (List.rev right))
          end

    fun splitr p l =
          let
            fun splitrF (e, (left, right, continue)) =
                  if continue andalso p e then
                    (left, e::right, true)
                  else
                    (e::left, right, false)

            val (left, right, _) =
                  foldr splitrF (nil, nil, true) l
          in
            (fromList left,
             fromList right)
          end

    fun splitri p l =
          let
            fun splitriF (i, e, (left, right, continue)) =
                  if continue andalso p (i, e) then
                    (left, e::right, true)
                  else
                    (e::left, right, false)

            val (left, right, _) =
                  foldri splitriF (nil, nil, true) l
                    
          in
            (fromList left,
             fromList right)
          end

    (* The take and drop functions can always be implemented by split. *)
    fun takel p l =
          #1 (splitl p l)

    fun takeli p l =
          #1 (splitli p l)

    fun taker p l =
          #2 (splitr p l)

    fun takeri p l =
          #2 (splitri p l)
    
    fun dropl p l =
          #2 (splitl p l)

    fun dropli p l =
          #2 (splitli p l)

    fun dropr p l =
          #1 (splitr p l)

    fun dropri p l =
          #1 (splitri p l)
  end

structure CharVector = String